/* Tsinghuaeolus soccer client for Robocup2001
 *
 * Computer Science Department
 * State Key Lab of Intelligent Technology and Systems
 * Tsinghua University
 * 
 * Copyright (C) 2001 Tsinghuaeolus
 * Team Leader : Lishi
 * Team memebers : ChenJiang, YaoJinyi, CaiYunpeng
 *
 * We thank CMU for CMU98 was used for reference in our initial stage.
 *
 * You may copy and distribute this program freely as long as you retain this notice.
 * If you make any changes or have any comments we would appreciate a message to yjy01@mails.tsinghua.edu.cn.
 * For more information, please see http://www.lits.tsinghua.edu.cn/RoboCup
 */

#include "stdafx.h"
#include <math.h>

#include "Command.h"
#include "utils.h"
#include "Global.h"


/*******************    Commmand   *******************************/
void Command::move(float X,float Y){//move command
	x = X;
	y = Y;
	type = CMD_move;
	time = situation.CurrentTime;
}
  
void Command::dash(float pow){//dash command
	power = pow;
	type = CMD_dash;
	time = situation.CurrentTime;
}

void Command::turn(float Moment){
	angle = Moment;
	type = CMD_turn;
	time = situation.CurrentTime;
}

void Command::kick(float pow,float ang){ //kick command
	power = pow;
	angle = ang;
	type = CMD_kick;
	time = situation.CurrentTime;
}

void Command::seize(float ang){ //seize command
	if ( !CP_goalie ) my_error("Only goalies can catch");
	if ( situation.playmode == PM_Before_Kick_Off ) return;

	type=CMD_catch;
	angle=ang;
	time = situation.CurrentTime;
}

void Command::say(char* msg){
	sprintf(cmdbuf,"(say %s)",msg);
	type = CMD_say;
	time = situation.CurrentTime;
}

void Command::turn_neck(float ang){
	angle = ang;
	pose_limitation(angle, SP_min_neck_moment, SP_max_neck_moment);
  
	type = CMD_turn_neck;
	time = situation.CurrentTime;
}

void Command::sense_body(){
	type = CMD_sense_body;
	time = situation.CurrentTime;
}

void Command::change_view(){
	type = CMD_change_view;
	time = situation.CurrentTime;
}

void Command::disconnect(){
	type = CMD_bye;
	time = situation.CurrentTime;
}

void Command::stay(){
	type = CMD_stay;
	time = situation.CurrentTime;
}

void Command::donothing(){
	type = CMD_donothing;
	time = situation.CurrentTime;
}

bool Command::valid(){
	return (bool) (type != CMD_none);
}

bool Command::valid(Time time){
	 return (bool) ((type != CMD_none && this->time == time)||
		 ((type == CMD_move) && (this->time == -1)));
}

/***********************    Actions  **********************************/
Actions::Actions(){
	latest = 0;
	kicks = 0;
	dashes = 0;
	turns = 0;
	turn_necks = 0;
	says = 0;
	lastcatchtime = -5;
}

/************ send initial message *******/
void Actions::send_init_message(){//send initializing message	
	if (IP_reconnect)
		sprintf(cmdbuf, "(reconnect %s %d)", CP_TeamName, IP_reconnect);
	else if ( CP_goalie && SP_version >= 4.00){
		sprintf(cmdbuf, "(init %s (version %.2f) (goalie))", CP_TeamName, SP_version);
	}
	else
		sprintf(cmdbuf, "(init %s (version %.2f))", CP_TeamName, SP_version);
	send_message(cmdbuf,sock);
}

/*******   coach commands  ********************/
void Actions::change_mode(PlayMode playmode){
	switch(playmode){
	case PM_No_Mode: return;
	case PM_Before_Kick_Off: 
		sprintf(cmdbuf,"(before_kick_off)");break;
	case PM_My_Kick_Off:
		sprintf(cmdbuf,"(kick_off_%c)",situation.MySide);break;
	case PM_Their_Kick_Off:
		sprintf(cmdbuf,"(kick_off_%c)",situation.TheirSide);break;
	case PM_My_Kick_In:
		sprintf(cmdbuf,"(kick_in_%c)",situation.MySide);break;
	case PM_Their_Kick_In:
		sprintf(cmdbuf,"(kick_in_%c)",situation.TheirSide);break;
	case PM_My_Corner_Kick:
		sprintf(cmdbuf,"(corner_kick_%c)",situation.MySide);break;
	case PM_Their_Corner_Kick:
		sprintf(cmdbuf,"(corner_kick_%c)",situation.TheirSide);break;
	case PM_My_Goal_Kick:
		sprintf(cmdbuf,"(goal_kick_%c)",situation.MySide);break;
	case PM_Their_Goal_Kick:
		sprintf(cmdbuf,"(goal_kick_%c)",situation.TheirSide);break;
	case PM_My_Free_Kick:
		sprintf(cmdbuf,"(free_kick_%c)",situation.MySide);break;
	case PM_Their_Free_Kick:
		sprintf(cmdbuf,"(kick_in_%c)",situation.TheirSide);break;
	default:
		break;
	}
	send_message(cmdbuf, sock);
}

void Actions::move_ball(float X,float Y){
	sprintf(cmdbuf,"(move (ball) %.2f %.2f)",X,Y);
	send_message(cmdbuf, sock);
}

void Actions::move_player(int side,UNum no,float X,float Y){
	if (side == 0)
		sprintf(cmdbuf,"(move (%s) %.2f %.2f)",MyTeamName,X,Y);
	else if(side == 1)
		sprintf(cmdbuf,"(move (%s) %.2f %.2f)",TheirTeamName,X,Y);
	send_message(cmdbuf, sock);
}

/****** see if commands take effects  ******************/
void Actions::ActionsInvalidation(){
	turn_valid = Commanded(CMD_turn, situation.LastCycle());
	dash_valid = Commanded(CMD_dash, situation.LastCycle());
	turn_neck_valid = Commanded(CMD_turn_neck, situation.LastCycle());
	kick_valid = Commanded(CMD_kick, situation.LastCycle());
	say_valid = Commanded(CMD_say, situation.LastCycle());
	if (sensory.SensedInfKnown(situation.CurrentTime)){
		SenseBodyInfo& sensebody = sensory.GetSenseBodyInfo(situation.CurrentTime);
		if (sensebody.turns < turns){
			my_error("Miss %d turns",turns - sensebody.turns);
			DoLog(LOG_MISS,"%d turns", turns - sensebody.turns);
			turneffects.Data(situation.LastCycle()) = 0;
			turn_valid = false;
		}
		if (sensebody.dashes < dashes){
			my_error("Miss %d dashes",dashes - sensebody.dashes);
			DoLog(LOG_MISS,"%d dashs", dashes - sensebody.dashes);
			dasheffects.Data(situation.LastCycle()) = 0;
			usedpower.Data(situation.LastCycle()) = 0;
			dash_valid = false;
		}
		if (sensebody.kicks < kicks){
			my_error("Miss %d kicks",kicks - sensebody.kicks);
			DoLog(LOG_MISS,"%d kicks",kicks - sensebody.kicks);
			kickeffects.Data(situation.LastCycle()) = 0;
			usedpower.Data(situation.LastCycle()) = 0;
			kick_valid = false;
		}
		if (sensebody.turn_necks < turn_necks){
			my_error("Miss %d turn_necks",turn_necks - sensebody.turn_necks);
			DoLog(LOG_MISS,"%d turn_necks",turn_necks - sensebody.turn_necks);
			turn_neckeffects.Data(situation.LastCycle()) = 0;
			turn_neck_valid = false;
		}
		if (sensebody.says < says){
			say_valid = false;
			my_error("Miss %d says",says - sensebody.says);
			DoLog(LOG_MISS,"%d says",says - sensebody.says);
		}
		turns = sensebody.turns;
		dashes = sensebody.dashes;
		kicks = sensebody.kicks;
		says = sensebody.says;
		turn_necks = sensebody.turn_necks;
	}
}

/***************   commands are sent by this function  **************/
void Actions::send_action(){
	if (!commands[latest].valid(situation.CurrentTime)){
		my_error("Is this command valid? CMD_type : %d",commands[latest].type);
		return;
	}

	switch (commands[latest].type){
	case CMD_kick:
		sprintf(cmdbuf,"(kick %.2f %.2f)", commands[latest].power, commands[latest].angle);
		kicks++;
		kickeffects.Setdata(Polar2Vector(commands[latest].power * Kickrate(), NormalizeAngle(commands[latest].angle + Self.bodyfacing)), situation.CurrentTime);
		usedpower.Setdata((float)fabs(commands[latest].power) * (commands[latest].power > 0 ? 1.0f : CP_negativepower_factor), situation.CurrentTime);
		DoLog(LOG_SENDACTION,"kickeffect (%.2f %.2f)", KickEffect(situation.CurrentTime).x, KickEffect(situation.CurrentTime).y);
		break;
	case CMD_dash: 
		sprintf(cmdbuf,"(dash %.2f)", commands[latest].power);
		dashes++;
		dasheffects.Setdata(Polar2Vector(commands[latest].power * Dashrate(), Self.bodyfacing), situation.CurrentTime);
		usedpower.Setdata((float)fabs(commands[latest].power) * (commands[latest].power > 0 ? 1.0f : CP_negativepower_factor), situation.CurrentTime);
		DoLog(LOG_SENDACTION, "dasheffect (%.2f %.2f)", DashEffect(situation.CurrentTime).x, DashEffect(situation.CurrentTime).y);
		break;
	case CMD_turn: 
		sprintf(cmdbuf,"(turn %.2f)", commands[latest].angle);
		turns++;
		turneffects.Setdata(commands[latest].angle * Turnrate(), situation.CurrentTime);
		DoLog(LOG_SENDACTION, "turneffect %.2f", TurnEffect(situation.CurrentTime));
		break;
	case CMD_turn_neck: 
		sprintf(cmdbuf,"(turn_neck %.2f)", commands[latest].angle);
		turn_necks++;
		turn_neckeffects.Setdata(commands[latest].angle, situation.CurrentTime);
		DoLog(LOG_SENDACTION, "turnneckeffect %.2f", TurnNeckEffect(situation.CurrentTime));
		break;
	case CMD_catch:
		sprintf(cmdbuf,"(catch %.2f)", commands[latest].angle);
		lastcatchtime = situation.CurrentTime;
		DoLog(LOG_SENDACTION, "catch %.2f", commands[latest].angle);
		break;
	case CMD_move:
		sprintf(cmdbuf,"(move %.2f %.2f)",commands[latest].x, commands[latest].y);
		DoLog(LOG_SENDACTION, "move");
		break;
	case CMD_sense_body:
		sprintf(cmdbuf,"(sense_body)");
		DoLog(LOG_SENDACTION, "sense_body");
		break;
	case CMD_change_view:
		sprintf(cmdbuf, "(change_view %s %s)", width_string, qual_string);
		DoLog(LOG_SENDACTION, "changeview");
		watch.Changeview();
		break;
	case CMD_bye:
		sprintf(cmdbuf, "(bye)");
		DoLog(LOG_SENDACTION, "bye");
		break;
	case CMD_say:
		DoLog(LOG_SENDACTION, "say");
		break;
	default: 
		return;
	}
	send_message(cmdbuf, sock);
	//DoLog(LOG_SENDACTION, cmdbuf);
	latest = (latest + 1)% CP_num_stored_CMDs;
}

void Actions::turn(float Moment){
	if (Moment > SP_max_moment){
		//my_error("Should not turn that positive much!");
		Moment = SP_max_moment;
	}
	if (Moment < SP_min_moment){
		//my_error("Should not turn that negative much!");
		Moment = SP_min_moment;
	}
	commands[latest].turn(Moment);
	send_action();
}

void Actions::dash(float power){
	float max, min;
	if (IsCriticalAction){
		max = max_power;
		min = min_power;
	}
	else{
		max = effort_max_power;
		min = effort_min_power;
	}
	pose_limitation(power, min, max);
	
	commands[latest].dash(power);
	send_action();
}

void Actions::kick(float power, float angle){
	if (power > SP_max_power){
		//my_error("Should not kick that positive much!");
		power = SP_max_power;
	}

	if (power < 0){
		//my_error("Should not kick that negative much!");
		power = 0;
	}

	commands[latest].kick(power, NormalizeAngle(angle));
	send_action();
}

void Actions::seize(float angle){
	commands[latest].seize(NormalizeAngle(angle));
	send_action();
}

void Actions::say(char *msg){
	commands[latest].say(msg);
	send_action();
}

void Actions::move(Vector pos){
	move(pos.x, pos.y);
}

void Actions::move(float X,float Y){
	commands[latest].move(X,Y);
	send_action();
}

void Actions::turn_neck(float ang){
	commands[latest].turn_neck(ang);
	send_action();
}

void Actions::sense_body(){
    commands[latest].sense_body();
	send_action();
}

void Actions::change_view(VIEWWIDTH width,VIEWQUALITY quality){
	switch (quality){
	case VQ_High: sprintf(qual_string,"high"); break;
	case VQ_Low:  sprintf(qual_string,"low");  break;
	}
	switch (width){
	case VW_Narrow: sprintf(width_string,"narrow"); break;
	case VW_Normal: sprintf(width_string,"normal"); break;
	case VW_Wide:   sprintf(width_string,"wide");   break;
	} 
	sensory.NextViewWidth = width; 
	commands[latest].change_view();
	send_action();
}

void Actions::disconnect(){
	commands[latest].disconnect();
	send_action();
}

void Actions::execute(Command& command){
	switch(command.type){
	case CMD_turn:
		turn(command.angle);
		break;
	case CMD_dash:
		dash(command.power);
		break;
	case CMD_kick:
		kick(command.power, command.angle);
		break;
	case CMD_turn_neck:
		turn_neck(command.angle);
		break;
	case CMD_catch:
		seize(command.angle);
		break;
	case CMD_sense_body:
		sense_body();
		break;
	case CMD_bye:
		disconnect();
		break;
	case CMD_none:
	default:
		break;
	}
}

bool Actions::Commanded(CMDType type, Time time){
	for(int i = latest - CP_num_commands_at_one_cycle; i < latest; i++){
		if(commands[CmdIndex(i)].time == time && commands[CmdIndex(i)].type == type)
			return true;
	}
	return false;
}

bool Actions::Valid(CMDType type){
	switch(type){
	case CMD_turn:
		return turn_valid;
	case CMD_kick:      
		return kick_valid;
	case CMD_dash:      
		return dash_valid;
	case CMD_turn_neck: 
		return turn_neck_valid;
	case CMD_say:       
		return say_valid;
	default:            
		return false;
	}
	return false;
}

AngleDeg Actions::Max_TurnAng(){
	return SP_max_moment * Turnrate();
}

AngleDeg Actions::Min_TurnAng(){
	return SP_min_moment * Turnrate();
}

float Actions::Kickrate(){
	return float( SP_kick_power_rate * (1 - 0.25f * fabs(NormalizeAngle(ball.global_angle - Self.bodyfacing)) /180 - 0.25f * (ball.distance - SP_player_size - SP_ball_size) / SP_kickable_margin) * Self.effort);
}

float Actions::Dashrate(){
	return float( SP_dash_power_rate * Self.effort);
}

float Actions::Turnrate(float speed){
	return float( 1 / (1.0f + SP_inertia_moment * speed));
}

float Actions::Turnrate(){
	return float( 1 / (1.0f + SP_inertia_moment * Self.speed));
}

float Actions::Max_KickEffect(){
	if (ball.kickable())
		return SP_max_power * Kickrate();
	else
		return 0;
}

float Actions::Max_DashEffect(){
	return SP_max_power * Dashrate();
}

bool Actions::cancatch(Time time){
	return (time - lastcatchtime > SP_catch_ban_cycle)&&(situation.playmode != PM_My_Goal_Kick);
}

float Actions::Kickrate(Time time){
	return kickrates.Data(time);
}

float Actions::Turnrate(Time time){
	return turnrates.Data(time);
}

float Actions::Dashrate(Time time){
	return dashrates.Data(time);
}

Vector Actions::KickEffect(Time time){
	return kickeffects.IsDataKnown(time) ? kickeffects.Data(time) : 0;
}

Vector Actions::DashEffect(Time time){
	return dasheffects.IsDataKnown(time) ? dasheffects.Data(time) : 0;
}

AngleDeg Actions::TurnEffect(Time time){
	return turneffects.IsDataKnown(time) ? turneffects.Data(time) : 0;
}

AngleDeg Actions::TurnNeckEffect(Time time){
	return turn_neckeffects.IsDataKnown(time) ? turn_neckeffects.Data(time) : 0;
}

float Actions::UsedPower(Time time){
	return usedpower.IsDataKnown(time) ? usedpower.Data(time) : 0;
}

void Actions::SetCritical(){
	IsCriticalAction = true;
}

void Actions::PreProcess(){
	kickrates.Setdata(Kickrate(), situation.CurrentTime);
	turnrates.Setdata(Turnrate(), situation.CurrentTime);
	dashrates.Setdata(Dashrate(), situation.CurrentTime);
	kickeffects.Setdata(0, situation.CurrentTime);
	turneffects.Setdata(0, situation.CurrentTime);
	turn_neckeffects.Setdata(0, situation.CurrentTime);
	dasheffects.Setdata(0, situation.CurrentTime);
	IsCriticalAction = false;
	effort_max_power = Max(Min(SP_max_power, Self.stamina - SP_effort_dec_thr * SP_stamina_max), 0.0f);
	effort_min_power = -Max(Min((float)fabs(SP_min_power), (Self.stamina - SP_effort_dec_thr * SP_stamina_max) /2.0f), 0.0f);

	max_power = Min(SP_max_power, Self.stamina);
	min_power = -Min((float)fabs(SP_min_power), Self.stamina / 2.0f);
}

Vector Actions::SummarizeKickeffectPos(Time prev, Time last, float speed_decay){
	Vector result = 0;
	float alpha = Exp(speed_decay, float(last - prev));
	for(Time time = prev; time < last; time ++){
		result += KickEffect(time) * ( 1 - alpha) / (1 - speed_decay);
		alpha /= speed_decay;
	}
	return result;
}

Vector Actions::SummarizeKickeffectVel(Vector ivel, Time prev, Time last, float speed_decay, float max_speed){
	for(Time time = prev; time < last; time ++){
		ivel += KickEffect(time);
		if (ivel.mod() > max_speed)
			ivel = ivel / ivel.mod() * max_speed;
		ivel *= speed_decay;
	}
	return ivel;
}

Vector Actions::SummarizeDasheffectPos(Vector ivel, Time prev, Time last, float speed_decay, float max_speed){
	Vector moved_pos = 0.0f;
	for(Time time = prev; time < last; time ++){
		ivel += DashEffect(time);
		if (ivel.mod() > max_speed){
			ivel = ivel / ivel.mod() * max_speed;
		}
		moved_pos += ivel;
		ivel *= speed_decay;
	}
	return moved_pos;
}

Vector Actions::SummarizeDasheffectVel(Vector ivel, Time prev, Time last, float speed_decay, float max_speed){
	for(Time time = prev; time < last; time ++){		
		ivel += DashEffect(time);
		if (ivel.mod() > max_speed){
			ivel = ivel / ivel.mod() * max_speed;
		}
		ivel *= speed_decay;
	}
	return ivel;
}

AngleDeg Actions::SummarizeTurneffect(Time prev, Time last){
	float sum_ang = 0.0f;
	for(Time time = prev; time < last; time ++){
		sum_ang += TurnEffect(time);
	}
	return NormalizeAngle(sum_ang);
}

AngleDeg Actions::SummarizeTurnneckeffect(Time prev, Time last){
	float sum_ang = 0.0f;
	for(Time time = prev; time < last; time ++){
		sum_ang += TurnNeckEffect(time);
	}
	return NormalizeAngle(sum_ang);
}

float Actions::CalcKickRate(AngleDeg ang, float distance){
	return float(SP_kick_power_rate * (1 - 0.25f * fabs(NormalizeAngle(ang)) /180 
			- 0.25f * (distance - SP_player_size - SP_ball_size) / SP_kickable_area) * Self.effort);
}

void Actions::ResetKickEffect(Time time){
	kickeffects.Setdata(0, time);
}
