/* Tsinghuaeolus soccer client for Robocup2001
 *
 * Computer Science Department
 * State Key Lab of Intelligent Technology and Systems
 * Tsinghua University
 * 
 * Copyright (C) 2001 Tsinghuaeolus
 * Team Leader : Lishi
 * Team memebers : ChenJiang, YaoJinyi, CaiYunpeng
 *
 * We thank CMU for CMU98 was used for reference in our initial stage.
 *
 * You may copy and distribute this program freely as long as you retain this notice.
 * If you make any changes or have any comments we would appreciate a message to yjy01@mails.tsinghua.edu.cn.
 * For more information, please see http://www.lits.tsinghua.edu.cn/RoboCup
 */

#include "df_positioning.h"
#include "global.h"



/***********	DF_Arrangement	***********************/
/*      generate subtasks for defense scheme          */
DF_Arrangement::DF_Arrangement(UNum defender, UNum defendee){
	df_size =  Vector(CP_defense_area_length, CP_defense_area_width);
	valid = Arrange_DF_for(defender, defendee);
}

/*
  press : when the opponent carrying ball is ahead of you so that you cannot
  block, try dashing back hard to keep threatening him
*/
bool DF_Arrangement::getpressposition(UNum My,UNum Opp,Vector &presspos){  
	if (fieldinfo.IsBlocked(defendee))
	{
		presspos = ball.pos;
		return true;
	}
	
	Ray theircourse;
	Vector l_mark,r_mark;
	Vector sectionpost,theirtarget;
	Vector MyPos = MyPlayer(My).pos;
	Vector Opos = TheirPlayer(Opp).pos;
	if (MyPos.dist(fieldinfo.mygoal) > Opos.dist(fieldinfo.mygoal))
		Opos += Polar2Vector(TheirPlayer(Opp).ActiveRadius(situation.CurrentTime) + TheirPlayer(Opp).max_speed * 2,
		(fieldinfo.mygoal - TheirPlayer(Opp).pos).Angle());
	//assume opponent's movement
	
	Line splitline;
	
	Line me_Opp;
	me_Opp.LineFromTwoPoints(MyPos,Opos);
	Vector goalpnt = me_Opp.intersection(fieldinfo.SideLines[SL_Left]);
	if (fabs(goalpnt.y) > SP_goal_width/2 * 1.05f)
	{
		l_mark = fieldinfo.l_mygoalpost;
		r_mark = fieldinfo.r_mygoalpost;
	}
	else
	{
		if (goalpnt.y < 0)
		{
			l_mark = fieldinfo.mygoal;
			r_mark = fieldinfo.r_mygoalpost;
		}
		else
		{
			l_mark = fieldinfo.l_mygoalpost;
			r_mark - fieldinfo.mygoal;
		}
	}
	splitline.LineFromTwoPoints((MyPos + Opos)/2,(l_mark + r_mark) /2);
	if (splitline.HalfPlaneTest(Opos) == splitline.HalfPlaneTest(l_mark))
	{
		sectionpost = l_mark;
		theirtarget = r_mark;
	}
	else 
	{
		sectionpost = r_mark;
		theirtarget = l_mark;
	}
	theircourse.SetValues(Opos,(theirtarget - Opos).Angle());
	Line mycourse;
	sectionpost = (sectionpost + Opos) /2; 
	mycourse.LineFromTwoPoints(MyPos,sectionpost);
	if  (!(theircourse.intersection(mycourse,presspos) && fieldinfo.WithInField(presspos)))
	{
		DoLog(LOG_BUG,"no press point 1");
		return false;
	}
	if (presspos.dist(MyPos) *1.1 < presspos.dist(Opos) && AngleDif((presspos - MyPos).Angle(),(presspos - Opos).Angle()) > 60
		&& presspos.dist(Opos) > 2.5f)
	{
		splitline.LineFrompline(MyPos,Opos);
		if (!theircourse.intersection(splitline,presspos))
		{
			DoLog(LOG_BUG,"no press point 2");
			return false;
		}
	}
	//consideration of bodyfacing
	Ray SelfRay(Self.pos,Self.bodyfacing);
	Vector secpoint;
	float secdist;
	float pressdist = theircourse.DistanceFromOrigin(theirtarget);
	if (theircourse.intersection(SelfRay,secpoint))
	{
		secdist = secpoint.dist(presspos);
		if (secdist < 1.0f || secdist < 0.15f * pressdist)
			presspos = secpoint;
	}
	return true;
}

/*
	block: intercept opponent carrying ball
*/
bool DF_Arrangement::getblockposition(UNum My, UNum Opp, Vector &blockpos){ 
	if (!IsTheirPlayer(Opp) || !IsMyPlayer(My)) return false;
	float cycles = 0.0f;
	Vector opppos;
	if (situation.BallFree){
		opppos = TheirPlayer(Opp).IT_inf.IT_point;
		cycles = TheirPlayer(Opp).IT_inf.IT_cycles;
	}
	else{
		opppos = TheirPlayer(Opp).pos;
		cycles = 0.0f;
	}
	float my_block_speed = SP_player_speed_max;
	float their_break_speed = SP_player_speed_max; 
 	if (!getblockpoint(opppos, MyPlayer(My).pos, cycles, my_block_speed, their_break_speed, blockpos))
		return false;
	if (!fieldinfo.WithInField(blockpos)) return false;
	float dsens = fieldinfo.GetDefensiveSensitivity(blockpos);
	if ( dsens > 0.95f && dsens > 1.1f * MyPlayer(My).defsensitivity.Data(situation.CurrentTime))
		return false;
	return true;
}

bool DF_Arrangement::getblockpoint(Vector opp_IT_pos, Vector selfpos, float opp_IT_cycles, float myspeed, float oppspeed, Vector& blockpos){
	float predis = myspeed * opp_IT_cycles;

	float disleft = opp_IT_pos.dist(fieldinfo.l_mygoalpost);
	float ratio = disleft/(disleft + opp_IT_pos.dist(fieldinfo.r_mygoalpost));
	Vector opttarget = Vector(-SP_pitch_length/2, SP_goal_width * (ratio - 0.5f));
	Vector opponent_pos = (opttarget - opp_IT_pos);
	opponent_pos.Normalize();
	opponent_pos = opp_IT_pos + opponent_pos * CP_block_margin;

	Vector vec_me_opt = selfpos - opponent_pos ;
	float dis_me_opt = vec_me_opt.mod();

	if(dis_me_opt <= predis || oppspeed < 1e-2){
		blockpos = opponent_pos;
		return true;
	}
	
	float ratio_speed = myspeed / oppspeed;
	
	if(fabs(NormalizeAngle(vec_me_opt.Angle()-(opttarget-opponent_pos).Angle()))<10)
		return getsubpoint(opttarget,opponent_pos,selfpos,ratio_speed,predis,blockpos);

	Vector subpos[3];
	bool block_ok[3];
	block_ok[0] = getsubpoint(opttarget,opponent_pos,selfpos,ratio_speed,predis,subpos[0]);
	block_ok[1] = getsubpoint(fieldinfo.l_mygoalpost,opponent_pos,selfpos,ratio_speed,predis,subpos[1]);	
	block_ok[2] = getsubpoint(fieldinfo.r_mygoalpost,opponent_pos,selfpos,ratio_speed,predis,subpos[2]);
	for(int i=1;i<3;i++){
		if(subpos[i].x<subpos[0].x){
			subpos[0]=subpos[i];
			block_ok[0]=block_ok[i];
		}
	}
	blockpos=subpos[0];
	return block_ok[0];	
}

bool DF_Arrangement::getsubpoint(Vector opttarget, Vector opponent_pos, Vector selfpos, float ratio_speed,float predis, Vector& blockpos){
	Vector vec_me_opt = selfpos - opponent_pos ;
	float dis_me_opt = vec_me_opt.mod();

	float interangle = vec_me_opt.Angle()-(opttarget-opponent_pos).Angle();
	float tempf,tempc;
	float tempa = 1.0f - ratio_speed * ratio_speed;
	if(fabs(tempa)<1e-2){
		tempf = predis + dis_me_opt * Cos(interangle);

		if(tempf <= 1e-2){
			blockpos = (opponent_pos + opttarget)/2;
			return false;
		}

		tempc = 0.5f*(dis_me_opt * dis_me_opt - predis*predis) / tempf;
	}
	else{
		tempf = 2*predis*ratio_speed + 2*dis_me_opt * Cos(interangle);
		float tempz = tempf * tempf - 4 * tempa*(dis_me_opt*dis_me_opt-predis*predis);
		if(tempz < 0){
			blockpos = (opponent_pos+opttarget)/2;
			return false;
		}
		tempz = (float)sqrt(tempz);
		float tempx1 = 0.5f * (-tempz + tempf)/tempa;
		float tempx2 = 0.5f * (tempz + tempf)/tempa;
		if(tempx1 < 0 && tempx2 < 0){
			blockpos = (opponent_pos + opttarget)/2;
			return false;
		}
		else if(tempx1 > 0 && tempx2 > 0) tempc = Min(tempx1, tempx2);
		else tempc = Max(tempx1,tempx2);	
	}
	float odist = opponent_pos.dist(opttarget);

	if(tempc >= odist + 1e-2){
		 blockpos = (opponent_pos + opttarget)/2;
		 return false;
	}
	
	float ratio = tempc / odist;

	blockpos = (opponent_pos * (1-ratio) + opttarget * ratio);
	
	return true;
}

/*
	mark : defend on opponents without ball,staying on the way that more
	convenience to get ball
*/
bool DF_Arrangement::getmarkposition(UNum My, UNum Opp, Vector ballpos, Vector& markpos){ 
//here global numbering scheme is assumed
	if (!IsTheirPlayer(Opp) || !IsMyPlayer(My)) return false;
	Vector dir1, dir2;
	Vector OppPos = TheirPlayer(Opp).pos;
	Vector target = ballpos.y>0 ? fieldinfo.r_mygoalpost : fieldinfo.l_mygoalpost;
	float dispersion =TheirPlayer(Opp).ActiveRadius(situation.CurrentTime) + TheirPlayer(Opp).max_speed;
	if (AngleDif((target - OppPos).Angle(),(target - Self.pos).Angle()) > 90)
		dispersion += 2.0f;
	else
		dispersion += 1.0f;
	pose_limitation(dispersion,0.0f,5.0f);
	float sensfactor = motion.sens_pos_factor.GetOutput(2*(fieldinfo.GetDefensiveSensitivity(OppPos) -0.5f));
	pose_limitation(sensfactor,0,1);
	OppPos += Polar2Vector(dispersion* (1 - sensfactor),(target - OppPos).Angle());	

	dir1 = ballpos - OppPos;
	dir1.Normalize();
	dir2 = fieldinfo.mygoal - OppPos;
	dir2.Normalize();
	dir1 *= sensfactor;
	dir2 *= 1 - sensfactor;
	float	markdist = CP_mark_dist;
	if(MyPlayer(My).IsRoleKnown()){
	//if a player's aggressiveness is high, he may act more aggressively when marking the opponent.
	//for instance, he will stand just in front of the opponent
		if (MyPlayer(My).aggressiveness >= 0.7f){
			markpos = dir1;
		}
		else if (MyPlayer(My).aggressiveness >= 0.5f){
			markpos = dir1 * 0.8f + dir2 * 0.2f;
		}
		else{
			markpos = dir1 * 0.5f + dir2 * 0.5f;
		}
	}
	else{
		markpos = dir1;
	}

	markpos.Normalize();
	markpos *= markdist;
	//consider body facing
	Ray markRay(OppPos,markpos);
	Ray SelfRay(Self.pos,Self.bodyfacing);
	Vector secpoint;
	float secdist;
	float balldist = ballpos.dist(OppPos);
	if (markRay.intersection(SelfRay,secpoint))
	{
		secpoint -=OppPos;
		secdist = secpoint.dist(markpos);
		if (secdist < 0.5f || secdist < 0.1f * markdist)
			markpos = secpoint;
	}
	markpos += OppPos;
	return true;
}

void DF_Arrangement::setpriority(){ 
	switch(df_type){
	case DF_Block:
		  priority = blockpriority = getblockpriority();
		  break;
	case DF_Formation:
		  priority = formationpriority = getformationpriority();
		  break;
	case DF_Mark:
		  priority = markpriority = getmarkpriority();
		  break;
	case DF_Press:
		if (fieldinfo.IsBlocked(defendee))
			presspriority = getblockpriority();
		else
			presspriority = getpresspriority();
			blockpriority = getblockpriority();
		priority = blockpriority;
		break;
	default:
		break;
	}
}

float DF_Arrangement::QueryPriority(bool block_applied){ 
	if (df_type == DF_Press && block_applied)
	{
		return presspriority;
	}
	else
		return priority;
}
float DF_Arrangement::getformationpriority(){
	double input,output;
	pos_dist = df_point.dist(MyPlayer(defender).pos);
	deviation_dist = df_point.dist(fm_pos);
	input = fieldinfo.GetDefensiveSensitivity(fm_pos);
	motion.FM_Priority_Net.SimulateNet(&input,&output);
	return (float)output;
}

float DF_Arrangement::getblockpriority(){
	double input[3], output;
	pos_dist = TheirPlayer(defendee).pos.dist(MyPlayer(defender).pos);
	deviation_dist = Min(TheirPlayer(defendee).pos.dist(fm_pos), df_point.dist(fm_pos));
	input[0] = threat;
	input[1] = deviation_dist / CP_deviation_max;
	input[2] = pos_dist / CP_deviation_max;
	motion.Block_Net.SimulateNet(&input[0], &output);
	output = (output - 0.1) * 1.25;
	if	(MyPlayer(defender).IsForward())
			output =0.5 + (output + TheirPlayer(defendee).GetOffsensitivity())/4;
	return (float)output;

}

float DF_Arrangement::getpresspriority(){
	double input[3], output;
	input[0] = threat;
	pos_dist = TheirPlayer(defendee).pos.dist(MyPlayer(defender).pos);
	deviation_dist = Min(TheirPlayer(defendee).pos.dist(fm_pos),df_point.dist(fm_pos));
	input[1] = deviation_dist / CP_deviation_max;
	input[2] = pos_dist / CP_deviation_max;
	motion.Press_Net.SimulateNet(&input[0], &output);
	output = (output - 0.1) * 1.25;
	if	(MyPlayer(defender).IsForward())
			output =0.5 + (output + TheirPlayer(defendee).GetOffsensitivity() )/4;
	return (float)output;


}

float DF_Arrangement::getmarkpriority(){
	double input[3], output;
	input[0] = threat;
	pos_dist = df_point.dist(MyPlayer(defender).pos);
	deviation_dist = df_point.dist(fm_pos);
	input[1] = deviation_dist / CP_deviation_max;
	input[2] = pos_dist / CP_deviation_max;
	motion.Mark_Net.SimulateNet(&input[0], &output);
	return float((output - 0.1) * 1.25);
}
void DF_Arrangement::setdfinfo(){
	fm_pos = fm.GetDefensivePoint(MyPlayer(defender), situation.NextControlBallPos());
	if (df_type == DF_Formation)
		threat = fieldinfo.GetDefensiveSensitivity(df_point);
	else
		threat = fieldinfo.CalcThreat(defendee);
	setpriority();
}

bool DF_Arrangement::Arrange_DF_for(UNum defender, UNum defendee){
	if (!IsMyPlayer(defender) || !MyPlayer(defender).IsRoleKnown()) return false;
	this->defender = defender;
	this->defendee = defendee;
	if (defendee == DF_FORMATION_DEFENDEE){
	//Defend formation
		df_type = DF_Formation;
		return Arrange_DF_Formation_for(defender);
	}
	else{
		if (!IsTheirPlayer(defendee))
			return false;
		if (!Rect(MyPlayer(defender).pos, df_size).IsWithin(TheirPlayer(defendee).pos))
			return false;
		if (situation.IsControllerOrFutureController(situation.TheirSide, defendee)){
			df_type = DF_Block;
			 if (Arrange_DF_Block_for(defender, defendee))
			 {
				 return true;
			 }
			 else
			 {
				 df_type = DF_Press;
				 return Arrange_DF_Press_for(defender,defendee);
			 }
		}
		else{		
			df_type = DF_Mark;
			return Arrange_DF_Mark_for(defender, defendee);
		}
	}
	return false;
}

bool DF_Arrangement::Arrange_DF_Formation_for(UNum defender){
	if (!IsMyPlayer(defender) || !MyPlayer(defender).IsRoleKnown()) return false;
	df_point = fm.GetDefensivePoint(MyPlayer(defender).attackness, MyPlayer(defender).leftness, situation.NextControlBallPos());
	setdfinfo();
	return true;
}

bool DF_Arrangement::Arrange_DF_Press_for(UNum defender, UNum defendee){
	if (!getpressposition(defender,defendee,df_point))
		return false;
	setdfinfo();
	return true;
}

bool DF_Arrangement::Arrange_DF_Block_for(UNum defender, UNum defendee){
	if (!getblockposition(defender, defendee, df_point))
		return false;
	setdfinfo();
	return true;
}

bool DF_Arrangement::Arrange_DF_Mark_for(UNum defender, UNum defendee){
	if (!getmarkposition(defender, defendee, situation.NextControlBallPos(), df_point))
		return false;
	setdfinfo();
	return true;
}

bool DF_Arrangement::operator == (const DF_Arrangement& rhs){
	return this->priority == rhs.priority;
}

bool DF_Arrangement::operator > (const DF_Arrangement& rhs){
	return this->priority > rhs.priority;
}

bool DF_Arrangement::operator >= (const DF_Arrangement& rhs){
	return this->priority >= rhs.priority;
}

bool DF_Arrangement::operator < (const DF_Arrangement& rhs){
	return this->priority < rhs.priority;
}

bool DF_Arrangement::operator <= (const DF_Arrangement& rhs){
	return this->priority <= rhs.priority;
}

bool DF_Arrangement::IsMutex(const DF_Arrangement& arrangement){
	return (this->defender == arrangement.defender)
		||	(this->defendee == arrangement.defendee && df_type != DF_Formation);
}

/***********	Defensive positioning	***********************/
DF_positioning::DF_positioning(){
	num_arrangements = 0;
}

void DF_positioning::AddDF_Arrangement(const DF_Arrangement& arrangement){
	if (arrangement.valid && (arrangement.priority >= 0.4f ||arrangement.df_type == DF_Block)){
		Order_of_arrangements[num_arrangements] = num_arrangements;
		DoLog(LOG_POSITIONING, "Add arrangement defender %d defendee %d pos_dist %.2f threat %.2f deviation %.2f point (%.2f %.2f) priority %.2f", 
			arrangement.defender, arrangement.defendee, arrangement.pos_dist, 
			arrangement.threat,arrangement.deviation_dist,
			arrangement.df_point.x, arrangement.df_point.y,
			arrangement.priority);
		arrangements[num_arrangements++] = arrangement;	
	}
}



void DF_positioning::ResetBuffer(){
	num_arrangements = 0;
	max_nonmutexs = 0;
	num_scheme_members = 0;
	block_applied_count = 0;
}

DF_Arrangement& DF_positioning::Arrangement(int idx){
	idx = (idx % MAX_DF_ARRANGEMENTS) >= 0 ? 
		(idx % MAX_DF_ARRANGEMENTS) : (idx % MAX_DF_ARRANGEMENTS) + MAX_DF_ARRANGEMENTS;
	return arrangements[Order_of_arrangements[idx]];
}

void DF_positioning::SortArrangements(){
	for(int i = 0; i < num_arrangements - 1; i ++)
		for(int j = i + 1; j < num_arrangements; j ++){
			if (Arrangement(j) > Arrangement(i)){
				Swap(Order_of_arrangements[i], Order_of_arrangements[j]);
			}
		}
}

bool DF_positioning::Do_arranging(){
	SortArrangements();
	int i, j, nonmutexs_test, tmp;
//set mutexs(for a fast searching)
	max_nonmutexs = 0;
	nonmutexs_test = 0;
	for(i = 0; i < num_arrangements; i ++){
		if (!((tmp = 1 << Arrangement(i).defender) & nonmutexs_test)){
			max_nonmutexs ++;
			nonmutexs_test |= tmp;
		}
		num_mutexs[i] = 0;
	}

	for(i = 0; i < num_arrangements - 1; i ++){
		for(j = i + 1; j < num_arrangements; j ++){
			if(Arrangement(i).IsMutex(Arrangement(j))){
				mutexs[i][num_mutexs[i]++] = j;
				mutexs[j][num_mutexs[j]++] = i;
			}
		}
	}
	num_scheme_members = 0;
	block_applied_count = 0;
	max_priority_sum = 0.0f;
	Searchfor_Scheme(0.0f);
	if (max_priority_sum == 0.0f){
		DoLog(LOG_POSITIONING, "no solution!");
		return false;
	}
	else{
		for(i = 0;  i < max_nonmutexs; i ++){
			if (Arrangement(optimal_scheme[i]).df_type == DF_Formation){
				DoLog(LOG_POSITIONING, "%d fm with %.2f %d", Arrangement(optimal_scheme[i]).defender, Arrangement(optimal_scheme[i]).priority, optimal_scheme[i]);
			}
			else{
				DoLog(LOG_POSITIONING, "%d-%d with %.2f %d", Arrangement(optimal_scheme[i]).defender, Arrangement(optimal_scheme[i]).defendee, Arrangement(optimal_scheme[i]).priority, optimal_scheme[i]);
			}
		}
		return true;
	}
	return false;
}

void DF_positioning::Searchfor_Scheme(float priority_sum){
	int idx = num_scheme_members > 0 ? scheme_members[num_scheme_members - 1] + 1: 0;
	int i;
	if(num_scheme_members >= max_nonmutexs){
		return;
	}

	bool is_mutex_set[2 * SP_team_size];
	while(idx <= num_arrangements + num_scheme_members - max_nonmutexs){
		if (Arrangement(idx).valid){
			scheme_members[num_scheme_members] = idx;

			if (num_scheme_members + 1 == max_nonmutexs){
				priority_sum += Arrangement(idx).QueryPriority(block_applied_count > 0);
				if (priority_sum > max_priority_sum){
					max_priority_sum = priority_sum;
					for(i = 0; i < num_scheme_members + 1; i ++)
						optimal_scheme[i] = scheme_members[i];
				}
				return;
			}


			if (priority_sum < max_priority_sum){
				float test_priority = priority_sum;
				for(i = 0; i < max_nonmutexs - num_scheme_members; i ++){
					test_priority += Arrangement(idx+i).QueryPriority(block_applied_count > 0);
				}

				if (test_priority < max_priority_sum)
					return;
			}

			num_scheme_members ++;

			for(i = 0; i < num_mutexs[idx]; i ++){
				if (mutexs[idx][i] > idx && Arrangement(mutexs[idx][i]).valid){
					Arrangement(mutexs[idx][i]).valid = false;
					is_mutex_set[i] = true;
				}
				else{
					is_mutex_set[i] = false;
				}
			}

			if (Arrangement(idx).df_type == DF_Block){
				block_applied_count ++;
			}

			Searchfor_Scheme(priority_sum + Arrangement(idx).QueryPriority(block_applied_count > 0));

			for(i = 0; i < num_mutexs[idx]; i ++){
				if (is_mutex_set[i]){
					Arrangement(mutexs[idx][i]).valid = true;
				}
			}

			if (Arrangement(idx).df_type == DF_Block){
				block_applied_count --;
			}

			num_scheme_members --;
		}
		idx ++;
	}
	return;
}

bool DF_positioning::MakeDefenseDecision(){
	ResetBuffer();
	int num_my, num_opp, i, j;
	num_my  = motion.Num_MyVisiblePlayers();
	num_opp = motion.Num_TheirVisiblePlayers();
	if (!Self.IsForward()){
		for(j = 0; j < num_opp; j ++){
			if (situation.IsControllerOrFutureController(situation.TheirSide,motion.TheirPlayers_Close2Me[j]) && situation.gamestate == GS_Their_PlaceKick)
					continue;
			AddDF_Arrangement(DF_Arrangement(MyNumber,
				motion.TheirPlayers_Close2Me[j]));
		}
	}

	AddDF_Arrangement(DF_Arrangement(MyNumber, DF_FORMATION_DEFENDEE));
	if (!Self.IsForward()){
		for(i = 0; i < num_my; i ++){
			if (motion.MyPlayer_Close2Me(i).Is_goalie) continue;
			if (!motion.MyPlayer_Close2Me(i).IsRoleKnown()) continue;
			if (motion.MyPlayer_Close2Me(i).IsForward()) continue;
			for(j = 0; j < num_opp; j ++){
				if (situation.IsControllerOrFutureController(situation.TheirSide, motion.TheirPlayers_Close2Me[j]) && situation.gamestate == GS_Their_PlaceKick)
					continue;
				AddDF_Arrangement(DF_Arrangement(motion.MyPlayers_Close2Me[i],
					motion.TheirPlayers_Close2Me[j]));
			}
			AddDF_Arrangement(DF_Arrangement(motion.MyPlayers_Close2Me[i],
					DF_FORMATION_DEFENDEE));
		}
	}
	if (num_arrangements <= 0)
		return false;
	return Do_arranging();
}

DF_Arrangement& DF_positioning::GetAssignment(){

	for(int i =0; i < max_nonmutexs; i ++){
		if (Arrangement(optimal_scheme[i]).defender == MyNumber)
			return Arrangement(optimal_scheme[i]);
	}
	//if no arrangement concerning me found, just return the first arrangement
	//there should be a mistake
	Arrangement(0).valid = false;
	return Arrangement(0);
}

float DF_positioning::Getmarkpriority(double threat, double deviation, double dist){
//Export for visual stategy 
	double input[3];
	double output;
	input[0] = threat;
	input[1] =deviation / CP_deviation_max;
	input[2] = dist / CP_deviation_max;
	motion.Mark_Net.SimulateNet(&input[0], &output);
	return float((output - 0.1) * 1.25);
}
