/* Tsinghuaeolus soccer client for Robocup2001
 *
 * Computer Science Department
 * State Key Lab of Intelligent Technology and Systems
 * Tsinghua University
 * 
 * Copyright (C) 2001 Tsinghuaeolus
 * Team Leader : Lishi
 * Team memebers : ChenJiang, YaoJinyi, CaiYunpeng
 *
 * We thank CMU for CMU98 was used for reference in our initial stage.
 *
 * You may copy and distribute this program freely as long as you retain this notice.
 * If you make any changes or have any comments we would appreciate a message to yjy01@mails.tsinghua.edu.cn.
 * For more information, please see http://www.lits.tsinghua.edu.cn/RoboCup
 */

#ifndef _mediator
#define _mediator

#include "Geometry.h"
#include "Command.h"
#include "bpn.h"
class Mediator;

class PrioredCommand : public Command{
private:
	float priority;
	ActionType actiontype;
public:
	PrioredCommand();
	PrioredCommand(Command& cmd, ActionType actiontype, float priority){ *this = cmd; this->actiontype = actiontype; this->priority = priority;}
	void Reset(){Command::Reset(); actiontype = Action_none; priority = 0.0f;}
	void operator =(const Command& cmd){angle = cmd.angle; x = cmd.x; y = cmd.y; 
	time = cmd.time; type = cmd.type; power = cmd.power;}
	bool operator >(const PrioredCommand& pcmd){ return priority > pcmd.priority;}
	bool operator ==(const PrioredCommand& pcmd){ return priority == pcmd.priority;}
	bool operator <(const PrioredCommand& pcmd){ return priority < pcmd.priority;}
	friend class Mediator;
	float GetPriority(){return priority;}
	ActionType GetActionType(){return actiontype;}
	friend class Mediator;
};


class VisualRequest{
private:
	float priority;
	AngleDeg angle;
	AngleDeg angle_inf;
	VIEWWIDTH view_width;
	bool valid;
public:
	VisualRequest(float priority = NoPriority, AngleDeg angle =0.0f);
	bool operator >(VisualRequest& vreq){ return  angle < vreq.angle;}
	bool operator ==(VisualRequest& vreq){ return angle == vreq.angle;}
	bool operator < (VisualRequest& vreq){ return angle > vreq.angle;}
	AngleDeg RequestedAngle(){return angle;};
	AngleDeg RequestedAngle_();
	float RequestedPriority();
	friend class Mediator;
};

/*****************	Mediator	*****************************/
class Mediator{
private:
/****	Priored Queue	**************/	
	Ordered_Soft_Queue<VisualRequest, CP_max_visualrequests> vr_queue;
	Ordered_Soft_Queue<PrioredCommand, CP_max_actions> ac_queue;

/*  visual request */
	VisualRequest visualrequests[CP_max_visualrequests];
	int totalvisualrequests;
	TimedData<VIEWWIDTH> designated_viewwidth;
/*********************************/
	KT_Res pass_res;
/*********************************/
	ActionType lastactiontype;
	CMDType lastcmdtype;
	void feedback(const PrioredCommand& bestaction , int vr_idx);
	float priority_mapping(ActionType actiontype, float priority);
public:
	Mediator();

	void ResetBuffer();
	bool enroll(PrioredCommand& pcmd);
	bool enroll(Command& command, ActionType actiontype, float priority);
	bool enroll(float priority, AngleDeg angle);
	void SetViewWidth(VIEWWIDTH view_width);
	void DisableViewWidth();
	void SetVisualRequests(VisualRequest* requests, int& totalrequests, VIEWWIDTH view_width, int max_requests = CP_max_visualrequests);
	void SetVisualRequests();
	void mediation();
	ActionType LastActionType();
	CMDType LastCmdType();
	
	void Getbestaction(PrioredCommand& bestaction);

/***	For Visual Decision and Audio Decision	***/
	bool IsLastCycleIntercept;//It is last cycle of interception, please change view
	TimedData<bool> PassOut;
	
	friend class VisualSystem; 
};

#endif